package jflowchart.engine.common;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.ComparableValue;
import jflowchart.engine.arithmetics.ComparableVariable;
import jflowchart.engine.arithmetics.GraphNodeOperation;
import jflowchart.engine.arithmetics.InvalidOperatorException;
import jflowchart.engine.arithmetics.NumberExpression;
import jflowchart.engine.arithmetics.NumberOperand;
import jflowchart.engine.arithmetics.NumberValue;
import jflowchart.engine.arithmetics.NumberVariable;
import jflowchart.engine.arithmetics.Operand;
import jflowchart.engine.arithmetics.Variable;


public class Functions {
	
	private static Integer parseInteger(String val) {
		int v;
		try {
			v = Integer.parseInt(val);
		} catch (NumberFormatException e) {
			return null;
		}
		return new Integer(v);
	}
	
	private static Float parseFloat(String val) {
		float v;
		try {
			v = Float.parseFloat(val);
		} catch (NumberFormatException e) {
			return null;
		}
		return new Float(v);
	}
	
	public static NumberValue parseNumberValue(String val) {
		val = val.trim();
		Integer i = parseInteger(val);
		if (i == null) {
			Float f = parseFloat(val);
			if (f == null) return null;
			else return new NumberValue(f);
		} else return new NumberValue(i);
	}
	
	public static ComparableValue parseComparableValue(String val) {
		val = val.trim();
		ComparableValue ret = parseNumberValue(val);
		if ((ret == null) && (! val.equals(Const.EMPTY_STRING))) 
				ret = new ComparableValue(val);
		return ret;
	}
	
	public static NumberVariable newIntegerVariable(String name) {
		return new NumberVariable(Integer.class, name, new NumberValue(0));
	}
	
	public static NumberVariable newFloatVariable(String name) {
		return new NumberVariable(Float.class, name, new NumberValue((float)0));
	}
	
	public static ComparableVariable newStringVariable(String name) {
		return new ComparableVariable(String.class, name, new ComparableValue(""));
	}
	
	@SuppressWarnings("rawtypes")
	public static Variable findVariable(Class type, String name, LinkedList<Variable> vars) {
		Variable v;
		for (int i = 0; i < vars.size(); i++) {
			v = vars.get(i);
			if ((v.getDataType().equals(type)) && (v.getName().equals(name)))
				return v;
		}
		return null;
	}
	
	@SuppressWarnings("rawtypes")
	public static Class parseDataType(String dataType) {
		Class ret = null;
		if (dataType.equals(Const.TYPE_LABELS[Const.INTEGER_TYPE_LABEL_CODE]))
			ret = Integer.class;
		else if (dataType.equals(Const.TYPE_LABELS[Const.FLOAT_TYPE_LABEL_CODE]))
			ret = Float.class;
		else if (dataType.equals(Const.TYPE_LABELS[Const.STRING_TYPE_LABEL_CODE]))
			ret = String.class;
		return ret;
	}
	
	public static NumberVariable parseNumberVariable(String type, String name, LinkedList<Variable> vars) {
		NumberVariable ret = (NumberVariable)Functions.findVariable(Functions.parseDataType(type), name, vars);
		if (ret == null) {
			if (type.equals(Const.TYPE_LABELS[Const.INTEGER_TYPE_LABEL_CODE]))
				ret = newIntegerVariable(name);
			else if (type.equals(Const.TYPE_LABELS[Const.FLOAT_TYPE_LABEL_CODE]))
				ret = newFloatVariable(name);
		}
		return ret;
	}
	
	public static ComparableVariable parseComparableVariable(String type, String name, LinkedList<Variable> vars) {
		ComparableVariable ret = parseNumberVariable(type, name, vars);
		if (ret == null) {
			ret = (ComparableVariable)Functions.findVariable(Functions.parseDataType(type), name, vars);
			if (ret == null) ret = newStringVariable(name);
		}
		return ret;
	}
	
	public static int parseComparisonOperatorCode(String sign) throws InvalidOperatorException {
		for (int i = 0; i < Const.COMPARISON_OPERATION_SIGNS.length; i++)
			if (Const.COMPARISON_OPERATION_SIGNS[i].equals(sign))
				return i;
		throw new InvalidOperatorException();
	}
	
	public static int parseNumberOperatorCode(String sign) throws InvalidOperatorException {
		for (int i = 0; i < Const.NUMBER_OPERATION_SIGNS.length; i++)
			if (Const.NUMBER_OPERATION_SIGNS[i].equals(sign))
				return i;
		throw new InvalidOperatorException();
	}
	
	public static boolean integers(Number a, Number b) {
		return (a.getClass().equals(Integer.class) && a.getClass().equals(Integer.class));
	}
	
	public static NumberValue add(Number a, Number b) {
		if (Functions.integers(a, b)) return new NumberValue(a.intValue() + b.intValue());
		else return new NumberValue(a.floatValue() + b.floatValue());
	}
	
	public static NumberValue sub(Number a, Number b) {
		if (Functions.integers(a, b)) return new NumberValue(a.intValue() - b.intValue());
		else return new NumberValue(a.floatValue() - b.floatValue());
	}
	
	public static NumberValue mul(Number a, Number b) {
		if (Functions.integers(a, b)) return new NumberValue(a.intValue() * b.intValue());
		else return new NumberValue(a.floatValue() * b.floatValue());
	}
	
	public static NumberValue div(Number a, Number b) {
		if (Functions.integers(a, b)) return new NumberValue(a.intValue() / b.intValue());
		else return new NumberValue(a.floatValue() / b.floatValue());
	}
	
	public static boolean numberExpression(NumberOperand o) {
		return o.getClass().equals(NumberExpression.class);
	}
	
	public static LinkedList<ComparableVariable> getComparableVariables(LinkedList<Variable> vars) {
		LinkedList<ComparableVariable> ret = new LinkedList<ComparableVariable>();
		for (int i = 0; i < vars.size(); i++)
			if (vars.get(i).isComparable())
				ret.add((ComparableVariable)vars.get(i));
		return ret;
	}
	
	public static LinkedList<Variable> filterNonComparableVariables(LinkedList<Variable> vars) {
		LinkedList<Variable> ret = new LinkedList<Variable>();
		for (int i = 0; i < vars.size(); i++)
			if (vars.get(i).isComparable())
				ret.add(vars.get(i));
		return ret;
	}
	
	public static LinkedList<NumberVariable> getNumberVariables(LinkedList<Variable> vars) {
		LinkedList<NumberVariable> ret = new LinkedList<NumberVariable>();
		for (int i = 0; i < vars.size(); i++)
			if (vars.get(i).isNumber())
				ret.add((NumberVariable)vars.get(i));
		return ret;
	}
	
	public static LinkedList<Variable> filterNonNumberVariables(LinkedList<Variable> vars) {
		LinkedList<Variable> ret = new LinkedList<Variable>();
		for (int i = 0; i < vars.size(); i++)
			if (vars.get(i).isNumber())
				ret.add(vars.get(i));
		return ret;
	}
	
	public static String[] getVariableComboItems(LinkedList<Variable> vars, Language lang, boolean newVarItem) {
		int add, varsCnt = vars.size();
		if (newVarItem) add = 2;
		else add = 1;
		String[] ret = new String[varsCnt + add];
		ret[0] = lang.selectVariable + Const.ELLIPSIS;
		Variable v;
		for (int i = 0; i < varsCnt; i++) {
			v = vars.get(i);
			ret[i+1] = v.printType() + Const.COLON + Const.SPACE + v.print();
		}
		if (newVarItem) ret[varsCnt + 1] = lang.newVariable + Const.ELLIPSIS;
		return ret;
	}
	
	public static NumberExpression parseNumberExpression(String[] data, int pointer, GraphNodeOperation parent) {
		NumberExpression ret = new NumberExpression();
		ret.setParent(parent);
		ret.paste(data, pointer);
		return ret;
	}
	
	public static Operand parseOperand(String[] data, int pointer, LinkedList<Variable> vars, GraphNodeOperation parent) {
		Operand ret = null;
		if (data[pointer].equals(Const.COMPARABLE_VALUE_CODE))
			ret = parseComparableValue(data[pointer+1]);
		else if (data[pointer].equals(Const.NUMBER_VALUE_CODE))
			ret = parseNumberValue(data[pointer+1]);
		else if (data[pointer].equals(Const.COMPARABLE_VARIABLE_CODE))
			ret = parseComparableVariable(data[pointer+1], data[pointer+2], vars);
		else if (data[pointer].equals(Const.NUMBER_VARIABLE_CODE))
			ret = parseNumberVariable(data[pointer+1], data[pointer+2], vars);
		else if (data[pointer].equals(Const.NUMBER_EXPRESSION_CODE)) {
			ret = parseNumberExpression(data, pointer, parent);
		}
		return ret;
	}

}
