package jflowchart.engine.nodes;

import java.util.LinkedList;
import java.util.UUID;

import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.codegen.c.CGenerator;
import jflowchart.engine.common.Const;
import jflowchart.engine.runtime.Program;
import jflowchart.engine.xml.XMLWriter;
import jflowchart.engine.xml.XMLWriterException;

public class ConnectPoint {

	protected String id;
	
	protected int x;
	protected int y;
	
	private Node owner;
	protected boolean input;
	private boolean horizontal;
	protected LinkedList<ConnectPoint> opposites;
	
	private int offset;
	private boolean upgoing;
	
	protected LinkedList<PreparationNode> loopOrigins;
	
	private void init(boolean in, boolean hor, int offset, boolean upgoing) {
		id = UUID.randomUUID().toString();
		x = -1;
		y = -1;
		input = in;
		horizontal = hor;
		opposites = new LinkedList<ConnectPoint>();
		this.offset = offset;
		this.upgoing = upgoing;
		loopOrigins = new LinkedList<PreparationNode>();
	}
	
	public void setID(String ID) {
		id = ID;
	}
	
	public String getID() {
		return id;
	}
	
	private void setOwner(Node n) {
		owner = n;
	}
	
	public ConnectPoint(boolean in, boolean hor) {
		init(in, hor, Const.NODE_CONNECTION_OFFSET, true);
	}
	
	public ConnectPoint(Node n, boolean in, boolean hor, int offset, boolean upgoing) {
		init(in, hor, offset, upgoing);
		setOwner(n);
	}
	
	public ConnectPoint(Node n, boolean in, boolean hor, int offset) {
		this(n, in, hor, offset, false);
	}
	
	public void setPosition(int posX, int posY) {
		x = posX;
		y = posY;
	}
	
	public void addOpposite(ConnectPoint cp) {
		if (input) {
			for (PreparationNode n : cp.getLoopOrigins())
				addLoopOrigin(n);
		} else {
			for (PreparationNode n : loopOrigins)
				cp.addLoopOrigin(n);
		}
		opposites.add(cp);
	}
	
	public boolean hasOpposites() {
		return (opposites.size() != 0);
	}
	
	public void removeLoopOrigin(PreparationNode n) {
		loopOrigins.remove(n);
		if (input) owner.discardLoopOrigin(n);
		else for (ConnectPoint cp : opposites) 
				cp.removeLoopOrigin(n);
	}
	
	public boolean hasLoopOrigins() {
		return (loopOrigins.size() != 0);
	}
	
	public void removeOpposite(ConnectPoint cp) {
		opposites.remove(cp);
		if ((input) && (cp.hasLoopOrigins())) { 
			for (PreparationNode n : cp.getLoopOrigins())
				removeLoopOrigin(n);
		} else if ((!input) && hasLoopOrigins()) {
			for (PreparationNode n : loopOrigins)
				cp.removeLoopOrigin(n);
		}
	}
	
	public void addLoopOrigin(PreparationNode n) {
		if (! loopOrigins.contains(n)) loopOrigins.add(n);
		if (input) owner.addLoopOrigin(n);
		else for (ConnectPoint cp : opposites) 
				cp.addLoopOrigin(n);
	}
	
	public int getX() {
		return x;
	}
	
	public int getY() {
		return y;
	}
	
	public Node getOwner() {
		return owner;
	}
	
	public boolean isInput() {
		return input;
	}
	
	public LinkedList<ConnectPoint> getOpposites() {
		return opposites;
	}
	
	public boolean isHorizontal() {
		return horizontal;
	}
	
	public LinkedList<PreparationNode> getLoopOrigins() {
		return loopOrigins;
	}
	
	protected int[] computeOffsetPoint(int offsetVal) {
		int[] point = new int[2];
		if (horizontal) {
			point[0] = x + offsetVal;
			point[1] = y;
		} else {
			point[0] = x;
			point[1] = y + offsetVal;
		}
		return point;
	}
	
	public int[] getOffsetPoint() {
		return computeOffsetPoint(offset);
	}
	
	public int[] getOffsetPoint(int customOffset) {
		if (offset < 0) customOffset = -1 * customOffset;
		return computeOffsetPoint(customOffset);
	}
	
	public void inverseOffset() {
		offset *= -1;
	}
	
	public int getOffset() {
		return offset;
	}
	
	protected void setOffset(int val) {
		offset = val;
	}
	
	public boolean isUnder(ConnectPoint cp) {
		return ((y > cp.getY()) || upgoing);
	}
	
	public boolean rightNextTo(ConnectPoint cp) {
		if (owner == null) return true;
		else if (cp.isAssistant()) return true;
		else if (owner.getPositionX() <= cp.getOwner().getPositionX()) 
			return (((x <= cp.getOwner().getWestBorder() && cp.isHorizontal())
					|| (x <= cp.getOwner().getPositionX() && (! cp.isHorizontal()))) 
					|| upgoing);
		else return ((x >= cp.getOwner().getEastBorder() && cp.isHorizontal()) 
				|| (x >= cp.getOwner().getPositionX() && (! cp.isHorizontal())));
	}
	
	public boolean movable() {
		if (hasOpposites()) {
			boolean ret = true;
			for (ConnectPoint cp : opposites) {
				if (isInput()) ret = (ret && (isUnder(cp) && rightNextTo(cp)));
				else ret = (ret && (cp.isUnder(this) && cp.rightNextTo(this)));
			}
			return ret;
		} else return true;
	}
	
	public boolean isLoopConnectPoint() {
		return false;
	}
	
	public boolean connectable(ConnectPoint opposite) {
		if (opposite.isAssistant()) return true;
		else if (isInput() && isLoopConnectPoint() && (this.hasLoopOrigins() && opposite.hasLoopOrigins() && (! opposite.getLoopOrigins().equals(this.getLoopOrigins()))))
			return false;
		else return (isUnder(opposite) && rightNextTo(opposite));
	}
	
	public boolean isReachableEnd(Program p) {
		if (!hasOpposites()) return false;
		else {
			boolean ret = true;
			for (ConnectPoint cp : opposites) {
				if (cp.isAssistant()) 
					ret = (ret && cp.isReachableEnd(p));
				else ret = (ret && cp.getOwner().isReachableEnd(p));
			}
			return ret;
		}
	}
	
	public boolean hasInfiniteLoop(Program p, PreparationNode origin) {
		if (isInput()) return owner.hasInfiniteLoop(p, origin);
		else if (!hasOpposites()) return false;
		else {
			boolean ret = true;
			for (ConnectPoint cp : opposites)
				ret = (ret && cp.hasInfiniteLoop(p, origin));
			return ret;
		}
	}
	
	public Variable checkVariables(Program p) {
		if (isInput()) return getOwner().checkVariables(p);
		else if (!hasOpposites()) return null;
		else {
			Variable v = null;
			for (ConnectPoint cp : opposites) {
				v = cp.checkVariables(p);
				if (v != null) return v;
			}
			return v;
		}
	}
	
	public boolean isFromDecision() {
		if (isInput() && (!hasOpposites())) return false;
		else if (isInput()) {
			for (ConnectPoint cp : getOpposites())
				if (cp.isFromDecision())
					return true;
			return false;
		} else return getOwner().isFromDecision();
			
	}
	
	public boolean isAssistant() {
		return false;
	}
	
	public void removeAssistant(boolean in) {
	}
	
	public ConnectPoint getOrigin() {
		return null;
	}
	
	public boolean isReachable(ConnectPoint p, boolean in) {
		return (this.equals(p));
	}
	
	public boolean checkLoops(PreparationNode n) {
		if (!hasOpposites()) return true;
		if (input) return owner.checkLoops(n);
		else return opposites.element().checkLoops(n);
	}
	
	public void nextStep(Program p) {
		if (input) p.setCurrentNode(owner);
		else opposites.element().nextStep(p);
	}
	
	public void save(XMLWriter h) throws XMLWriterException {
		h.saveConnectPoint(id);
	}
	
	public void collectVariables(LinkedList<Variable> vars) {
		if (input) owner.collectVariables(vars);
		else opposites.element().collectVariables(vars);
	}
	
	public void generateCCode(CGenerator g) {
		if (input) owner.generateCCode(g);
		else opposites.element().generateCCode(g);
	}
	
}
