package jflowchart.engine.nodes;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.Comparison;
import jflowchart.engine.arithmetics.ConditionValue;
import jflowchart.engine.arithmetics.Operation;
import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.codegen.c.CGenerator;
import jflowchart.engine.common.Const;
import jflowchart.engine.runtime.Program;
import jflowchart.engine.xml.XMLWriter;
import jflowchart.engine.xml.XMLWriterException;

public class DecisionNode extends GeneralNode {
	
	protected Comparison operation;
	
	private void init() {
		operation = new Comparison();
		connectPoints[0] = new ConnectPoint(this, true, false, -Const.NODE_CONNECTION_OFFSET);
		connectPoints[1] = new ConnectPoint(this, false, true, -Const.NODE_CONNECTION_OFFSET);
		connectPoints[2] = new ConnectPoint(this, false, true, Const.NODE_CONNECTION_OFFSET);
		setConnectPoints();
	}
	
	protected DecisionNode(boolean ghost) {
		super(Const.DECISION_NODE_WIDTH, Const.DECISION_NODE_HEIGHT, 3, ghost);
		init();
	}
	
	protected DecisionNode(int posX, int posY, boolean ghost) {
		super(posX, posY, Const.DECISION_NODE_WIDTH, Const.DECISION_NODE_HEIGHT, 3, ghost);
		init();
	}
	
	public Operation getOperation() {
		return operation;
	}

	public String copy(String textData) {
		textData += Const.DECISION_CODE + Const.SPACE;
		textData = super.copy(textData) + Const.SPACE;
		textData += operation.copy();
		textData = separateClipboardData(textData);
		return textData;
	}
	
	@Override
	public boolean paste(String[] data) {
		if (! data[0].equals(Const.DECISION_CODE)) return false;
		else if (! super.paste(data)) return false;
		else if (operation.paste(data, 3) == Const.PASTE_ERROR) return false;
		else return operation.usable();
	}

	@Override
	public void setConnectPoints() {
		connectPoints[0].setPosition(positionX, getNorthBorder());
		connectPoints[1].setPosition(getWestBorder(), positionY);
		connectPoints[2].setPosition(getEastBorder(), positionY);
	}

	@Override
	public ConnectPoint connect(int x, int y) {
		if (isInside(x, y) && isBetween(y, this.getNorthBorder(), (this.getNorthBorder() + ((2 * (this.positionY-this.getNorthBorder())) / 3)))) 
			return connectPoints[0];
		else if (isInside(x, y) && isBetween(x, this.getWestBorder(), this.positionX))
			return connectPoints[1];
		else if (isInside(x, y) && isBetween(x, this.positionX, this.getEastBorder()))
			return connectPoints[2];
		else return null;
	}
	
	@Override
	public ConnectPoint connect(int x, int y, boolean input) {
		if (isInside(x, y)) {
			if (input) return connectPoints[0];
			else if (isBetween(x, this.getWestBorder(), this.positionX))
				return connectPoints[1];
			else return connectPoints[2];
		} else return null;
	}

	@Override
	public boolean isReachableEnd(Program p) {
		if (operation.isPassing() == ConditionValue.TRUE) return connectPoints[1].isReachableEnd(p);
		else if (operation.isContradiction() == ConditionValue.FALSE) return connectPoints[2].isReachableEnd(p);
		else return (connectPoints[1].isReachableEnd(p) && connectPoints[2].isReachableEnd(p));
	}

	@Override
	public void addLoopOrigin(PreparationNode n) {
		connectPoints[1].addLoopOrigin(n);
		connectPoints[2].addLoopOrigin(n);
	}

	@Override
	public boolean hasInfiniteLoop(Program p, PreparationNode origin) {
		if (operation.isPassing() == ConditionValue.TRUE) return connectPoints[1].hasInfiniteLoop(p, origin);
		else if (operation.isContradiction() == ConditionValue.FALSE) return connectPoints[2].hasInfiniteLoop(p, origin);
		else return (connectPoints[1].hasInfiniteLoop(p, origin) && connectPoints[2].hasInfiniteLoop(p, origin));
	}

	@Override
	public Variable checkVariables(Program p) {
		Variable v = operation.checkVariables(p);
		if (v != null) return v;
		else {
			v = connectPoints[1].checkVariables(p);
			if (v == null) v = connectPoints[2].checkVariables(p);
			return v;
		}
	}

	@Override
	public boolean isFromDecision() {
		return true;
	}

	@Override
	public void discardLoopOrigin(PreparationNode n) {
		connectPoints[1].removeLoopOrigin(n);
		connectPoints[2].removeLoopOrigin(n);
	}

	@Override
	public boolean checkLoops(PreparationNode origin) {
		if (operation.isPassing() == ConditionValue.TRUE) return connectPoints[1].checkLoops(origin);
		else if (operation.isContradiction() == ConditionValue.FALSE) return connectPoints[2].checkLoops(origin);
		else return (connectPoints[1].checkLoops(origin) && connectPoints[2].checkLoops(origin));
	}

	@Override
	public void execute(Program p) {
		if (hasBreakpoint && p.isRunning()) p.pause();
		else {
			if (operation.evaluate())
				connectPoints[1].nextStep(p);
			else connectPoints[2].nextStep(p);
		}
	}

	@Override
	public void save(XMLWriter h) throws XMLWriterException {
		h.openNodeElem(Const.DECISION_CODE, positionX, positionY);
		h.saveConnectPoints(connectPoints);
		operation.save(h);
		h.closeElem();
	}

	@Override
	public void generateCCode(CGenerator g) {
		g.newLine("if (" + operation.print() + ") {");
		g.incrIndent();
		connectPoints[1].generateCCode(g);
		g.decrIndent();
		g.newLine("} else {");
		g.incrIndent();
		connectPoints[2].generateCCode(g);
		g.decrIndent();
		g.newLine("}");
	}

	@Override
	public void collectVariables(LinkedList<Variable> vars) {
		operation.collectVariables(vars);
		connectPoints[1].collectVariables(vars);
		connectPoints[2].collectVariables(vars);
	}
	
}
