package jflowchart.engine.nodes;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.GraphNodeOperation;
import jflowchart.engine.common.Const;
import jflowchart.engine.xml.XMLReaderException;

public abstract class GeneralNode implements Node {
	
	protected int positionX;
	protected int positionY;
	
	protected int width;
	protected int height;
	
	protected boolean ghost;
	protected boolean selected;
	protected boolean moving;
	
	protected GraphNodeOperation operation;
	
	protected ConnectPoint[] connectPoints;
	
	protected boolean hasBreakpoint;
	protected boolean isActive;
	
	private void setSize(int w, int h) {
		width = w;
		height = h;
	}
	
	private void init(int x, int y, int w, int h, int connectPts, boolean ghost) {
		
		connectPoints = new ConnectPoint[connectPts];
		
		positionX = x;
		positionY = y;
		setSize(w, h);
		
		this.ghost = ghost;
		this.selected = false;
		this.moving = false;
		
		hasBreakpoint = false;
	
	}
	
	protected GeneralNode(int w, int h, int connectPts, boolean ghost) {
		init(-1, -1, w, h, connectPts, ghost);
		operation = null;
	}
	
	protected GeneralNode(int posX, int posY, int w, int h, int connectPts, boolean ghost) {
		init(posX, posY, w, h, connectPts, ghost);
		operation = null;
	}
	
	public void setPosition(int x, int y) {	
		positionX = x;
		positionY = y;
		setConnectPoints();
	}
	
	public void setSelected(boolean val) {
		selected = val;
	}
	
	public void setGhost(boolean val) {
		ghost = val;
	}
	
	public void setMoving(boolean val) {
		moving = val;
	}
	
	private boolean movable() {
		for (ConnectPoint cp : connectPoints) {
			if (! cp.movable()) return false;
		}
		return (positionX >= 0) && (positionY >= 0);
	}
	
	public void move(int dx, int dy) {
		
		positionX += dx;
		positionY += dy;
		setConnectPoints();
		
		if (! movable()) {
			positionX -= dx;
			positionY -= dy;
			setConnectPoints();
		}
	
	}
	
	public int getPositionX() {
		return positionX;
	}
	
	public int getPositionY() {
		return positionY;
	}
	
	public int getWidth() {
		return width;
	}
	
	public int getHeight() {
		return height;
	}
	
	public int getWestBorder() {
		return positionX - Math.round(width/2);
	}
	
	public int getEastBorder() {
		return positionX + Math.round(width/2);
	}
	
	public int getNorthBorder() {
		return positionY - Math.round(height/2);
	}
	
	public int getSouthBorder() {
		return positionY + Math.round(height/2);
	}
	
	protected boolean isBetween(int x, int left, int right) {
		return ((left < x) && (x < right));
	}
	
	public boolean isInside(int x, int y) {
		return isBetween(x, this.getWestBorder(), this.getEastBorder()) 
				&& isBetween(y, this.getNorthBorder(), this.getSouthBorder());
	}
	
	public boolean isInRectangle(int x, int y, int width, int height) {
		return (((x < positionX) && (positionX < x + width)) 
				&& ((y < positionY) && (positionY < y + height)));
	}
	
	public boolean isGhost() {
		return ghost;
	}
	
	public boolean isSelected() {
		return selected;
	}
	
	public boolean isMoving() {
		return moving;
	}
	
	public String copy(String textData) {
		textData += String.valueOf(positionX) + Const.SPACE;
		textData += String.valueOf(positionY);
		return textData;
	}
	
	public boolean paste(String[] data) {
		try {
			positionX = Integer.parseInt(data[1]);
			positionY = Integer.parseInt(data[2]);
		} catch(Exception e) {
			return false;
		}
		return true;
	}
	
	protected String separateClipboardData(String textData) {
		textData += Const.SPACE + Const.CLIPBOARD_TEXTDATA_SEPARATOR + Const.SPACE;
		return textData;
	}
	
	public ConnectPoint[] getConnectPoints() {
		return connectPoints;
	}
	
	public boolean isStart() {
		return false;
	}
	
	@Override
	public void toggleBreakpoint() {
		hasBreakpoint = (!hasBreakpoint);
	}

	@Override
	public boolean hasBreakpoint() {
		return hasBreakpoint;
	}
	
	@Override
	public void setActive(boolean val) {
		isActive = val;
	}

	@Override
	public boolean isActive() {
		return isActive;
	}
	
	public void setConnectPointIDs(LinkedList<String> IDs) throws XMLReaderException {
		if (IDs.size() != connectPoints.length)
			throw new XMLReaderException("Invalid count of connect points");
		else for (int i = 0; i < connectPoints.length; i++)
			connectPoints[i].setID(IDs.get(i));
	}
	
	@Override
	public ConnectPoint getConnectPointByID(String id) {
		for (ConnectPoint cp : connectPoints)
			if (cp.getID().equals(id))
				return cp;
		return null;
	}

}
