package jflowchart.engine.nodes;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.Assignment;
import jflowchart.engine.arithmetics.Comparison;
import jflowchart.engine.arithmetics.ConditionValue;
import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.codegen.c.CGenerator;
import jflowchart.engine.common.Const;
import jflowchart.engine.runtime.Program;
import jflowchart.engine.xml.XMLWriter;
import jflowchart.engine.xml.XMLWriterException;

public class PreparationNode extends GeneralNode {

	protected Comparison condition;
	protected Assignment operation;
	
	private boolean secondPass;
	private boolean noInfiniteLoop;
	private boolean looping;
	
	private void init() {
		
		condition = new Comparison();
		operation = new Assignment();
		
		secondPass = false;
		noInfiniteLoop = false;
		looping = false;
		
		connectPoints[0] = new ConnectPoint(this, true, false, -Const.NODE_CONNECTION_OFFSET);
		connectPoints[1] = new LoopConnectPoint(this, true, true, -Const.NODE_CONNECTION_OFFSET*2);
		connectPoints[2] = new ConnectPoint(this, false, true, Const.NODE_CONNECTION_OFFSET);
		connectPoints[3] = new LoopConnectPoint(this, false, false, Const.NODE_CONNECTION_OFFSET);
		
		setConnectPoints();
		
	}
	
	protected PreparationNode(boolean ghost) {
		super(Const.PREPARATION_NODE_WIDTH, Const.PREPARATION_NODE_HEIGHT, 4, ghost);
		init();
	}
	
	protected PreparationNode(int posX, int posY, boolean ghost) {
		super(posX, posY, Const.PREPARATION_NODE_WIDTH, Const.PREPARATION_NODE_HEIGHT, 4, ghost);
		init();
	}
	
	public Comparison getCondition() {
		return condition;
	}

	public String copy(String textData) {
		textData += Const.PREPARATION_CODE + Const.SPACE;
		textData = super.copy(textData) + Const.SPACE;
		textData += condition.copy() + Const.SPACE;
		textData += operation.copy();
		textData = separateClipboardData(textData);
		return textData;
	}
	
	@Override
	public boolean paste(String[] data) {
		if (! data[0].equals(Const.PREPARATION_CODE)) return false;
		else if (! super.paste(data)) return false;
		int ptr = condition.paste(data, 3);
		if (ptr == Const.PASTE_ERROR) return false;
		else {
			ptr = operation.paste(data, ptr);
			if (ptr == Const.PASTE_ERROR) return false;
			else return (condition.usable() && operation.usable());
		}
	}

	@Override
	public Assignment getOperation() {
		return operation;
	}

	@Override
	public void setConnectPoints() {
		connectPoints[0].setPosition(positionX, getNorthBorder());
		connectPoints[1].setPosition(getWestBorder(), positionY);
		connectPoints[2].setPosition(getEastBorder(), positionY);
		connectPoints[3].setPosition(positionX, getSouthBorder());
	}
	
	private int nSixth(int n) {
		return this.getWestBorder() + (((this.getEastBorder() - this.getWestBorder()) * n) / 6);
	}
	
	private boolean hasLoop() {
		return (connectPoints[1].hasLoopOrigins() && connectPoints[1].getLoopOrigins().contains(this));
	}
	
	public boolean checkLoops(PreparationNode n) {
		if (this.equals(n)) return true;
		else if (!hasLoop())
			return false;
		else {
			boolean inside = connectPoints[3].checkLoops(this);
			if (!inside) return false;
			else return connectPoints[2].checkLoops(n);
		}
	}

	@Override
	public ConnectPoint connect(int x, int y) {
		if (isInside(x, y) && isBetween(x, nSixth(1), nSixth(5)) && isBetween(y, this.getNorthBorder(), this.positionY))
			return connectPoints[0];
		else if (isInside(x, y) && isBetween(x, this.getWestBorder(), nSixth(1)))
			return connectPoints[1];
		else if (isInside(x, y) && isBetween(x, nSixth(5), this.getEastBorder()))
			return connectPoints[2];
		else if (isInside(x, y) && isBetween(x, nSixth(1), nSixth(5)) && isBetween(y, this.positionY, this.getSouthBorder()))
			return connectPoints[3];
		else return null;
	}
	
	@Override
	public ConnectPoint connect(int x, int y, boolean input) {
		if (isInside(x, y)) {
			if (input) {
				if (isBetween(x, this.getWestBorder(), nSixth(1)))
					return connectPoints[1];
				else return connectPoints[0];
			} else {
				if (isBetween(x, nSixth(5), this.getEastBorder()))
					return connectPoints[2];
				else return connectPoints[3];
			}
		} else return null;
	}

	@Override
	public boolean isReachableEnd(Program p) {
		recordIterator(p);
		if (hasLoop() && (!checkLoops(null))) return false;
		if (hasLoop()) return connectPoints[2].isReachableEnd(p);
		else if (condition.isPassing() == ConditionValue.TRUE) return connectPoints[3].isReachableEnd(p);
		else if (condition.isPassing() == ConditionValue.FALSE) return connectPoints[2].isReachableEnd(p);
		else return (connectPoints[3].isReachableEnd(p) && connectPoints[2].isReachableEnd(p));
	}

	@Override
	public void addLoopOrigin(PreparationNode n) {
		if ((n != null) && (!n.equals(this))) connectPoints[2].addLoopOrigin(n);
	}
	
	private boolean hasTautology() {
		return (condition.getFirstOperand().hasFixValue() 
				&& condition.getSecondOperand().hasFixValue());
	}

	@Override
	public boolean hasInfiniteLoop(Program p, PreparationNode origin) {
		if (this.equals(origin)) return false;
		else if (noInfiniteLoop) return false;
		else if (hasLoop() && (condition.isPassing() == ConditionValue.TRUE) && hasTautology()) return true;
		else {
			noInfiniteLoop = true;
			recordIterator(p);
			boolean inside = connectPoints[3].hasInfiniteLoop(p, this);
			noInfiniteLoop = false;
			if (!inside) return connectPoints[2].hasInfiniteLoop(p, this);
			else return inside;
		}
	}
	
	private Variable checkBreak(Program p) {
		secondPass = false;
		return connectPoints[2].checkVariables(p);
	}

	@Override
	public Variable checkVariables(Program p) {
		if (! secondPass) {
			Variable v = condition.checkVariables(p);
			if (v != null) return v;
			else {
				v = operation.checkVariables(p);
				if (v != null) return v;
				else {
					secondPass = true;
					recordIterator(p);
					v = connectPoints[3].checkVariables(p);
					if ((v == null) && (secondPass)) return checkBreak(p);
					else {
						secondPass = false;
						return v;
					}
				}
			}
		} else return checkBreak(p);
	}

	@Override
	public boolean isFromDecision() {
		return connectPoints[0].isFromDecision();
	}

	@Override
	public void discardLoopOrigin(PreparationNode n) {
		connectPoints[2].removeLoopOrigin(n);
	}
	
	private void recordIterator(Program p) {
		Variable v = operation.getFirstOperand();
		p.setIterating(v);
	}
	
	private void performStepOperation(Program p) {
		p.setValue(operation.getFirstOperand(), operation.getSecondOperand().getValue());
		p.updateMonitor(operation.getFirstOperand());
		looping = false;
	}
	
	public void checkLooping() {
		looping = true;
	}

	@Override
	public void execute(Program p) {
		if (hasBreakpoint && p.isRunning()) p.pause();
		else {
			if (looping) performStepOperation(p);
			if (condition.evaluate())
				connectPoints[3].nextStep(p);
			else connectPoints[2].nextStep(p);
		}
	}
	
	@Override
	public void save(XMLWriter h) throws XMLWriterException {
		h.openNodeElem(Const.PREPARATION_CODE, positionX, positionY);
		h.saveConnectPoints(connectPoints);
		condition.save(h);
		operation.save(h);
		h.closeElem();
	}

	@Override
	public void generateCCode(CGenerator g) {
		g.newLine("while (" + condition.print() + ") {");
		g.incrIndent();
		connectPoints[3].generateCCode(g);
		g.newLine(operation.print() + Const.SEMICOLON);
		g.decrIndent();
		g.newLine("}");
		connectPoints[2].generateCCode(g);
	}

	@Override
	public void collectVariables(LinkedList<Variable> vars) {
		condition.collectVariables(vars);
		operation.collectVariables(vars);
		connectPoints[3].collectVariables(vars);
		connectPoints[2].collectVariables(vars);
	}

}
