package jflowchart.engine.nodes;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.Assignment;
import jflowchart.engine.arithmetics.Operand;
import jflowchart.engine.arithmetics.Value;
import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.codegen.c.CGenerator;
import jflowchart.engine.common.Const;
import jflowchart.engine.runtime.Program;
import jflowchart.engine.xml.XMLWriter;
import jflowchart.engine.xml.XMLWriterException;

public class ProcessNode extends GeneralNode {

	protected Assignment operation;
	
	private void init() {
		operation = new Assignment();
		connectPoints[0] = new ConnectPoint(this, true, false, -Const.NODE_CONNECTION_OFFSET);
		connectPoints[1] = new ConnectPoint(this, false, false, Const.NODE_CONNECTION_OFFSET);
		setConnectPoints();
	}
	
	protected ProcessNode(boolean ghost) {
		super(Const.PROCESS_NODE_WIDTH, Const.PROCESS_NODE_HEIGHT, 2, ghost);
		init();
	}
	
	protected ProcessNode(int posX, int posY, boolean ghost) {
		super(posX, posY, Const.PROCESS_NODE_WIDTH, Const.PROCESS_NODE_HEIGHT, 2, ghost);
		init();
	}
	
	public Assignment getOperation() {
		return operation;
	}

	public String copy(String textData) {
		textData += Const.PROCESS_CODE + Const.SPACE;
		textData = super.copy(textData) + Const.SPACE;
		textData += operation.copy();
		textData = separateClipboardData(textData);
		return textData;
	}
	
	@Override
	public boolean paste(String[] data) {
		if (! data[0].equals(Const.PROCESS_CODE)) return false;
		else if (! super.paste(data)) return false;
		else if (operation.paste(data, 3) == Const.PASTE_ERROR) return false;
		else return operation.usable();
	}

	@Override
	public void setConnectPoints() {
		connectPoints[0].setPosition(positionX, getNorthBorder());
		connectPoints[1].setPosition(positionX, getSouthBorder());
	}

	@Override
	public ConnectPoint connect(int x, int y) {
		if (isInside(x, y) && isBetween(y, this.getNorthBorder(), this.positionY)) return connectPoints[0];
		else if (isInside(x, y) && isBetween(y, this.positionY, this.getSouthBorder())) return connectPoints[1];
		return null;
	}
	
	@Override
	public ConnectPoint connect(int x, int y, boolean input) {
		if (isInside(x, y)) {
			if (input) return connectPoints[0];
			else return connectPoints[1];
		} else return null;
	}
	
	public boolean isReachableEnd(Program p) {
		recordVariable(p);
		return connectPoints[1].isReachableEnd(p);
	}

	@Override
	public void addLoopOrigin(PreparationNode n) {
		connectPoints[1].addLoopOrigin(n);
	}
	
	@Override
	public boolean hasInfiniteLoop(Program p, PreparationNode origin) {
		recordVariable(p);
		return connectPoints[1].hasInfiniteLoop(p, origin);
	}
	
	private void recordVariable(Program p) {
		Variable left = operation.getFirstOperand();
		p.addVariable(left);
		Operand right = operation.getSecondOperand();
		if (right.isUserSet()) p.setUserSet(left);
		else if (right.isIterating()) {
			p.setValue(left, right.getValue());
			p.setIterating(left);
		}
		else p.setValue(left, right.getValue());
	}

	@Override
	public Variable checkVariables(Program p) {
		Variable v = operation.checkVariables(p);
		if (v == null) {
			recordVariable(p);
			return connectPoints[1].checkVariables(p);
		} else return v;
	}

	@Override
	public boolean isFromDecision() {
		return connectPoints[0].isFromDecision();
	}

	@Override
	public void discardLoopOrigin(PreparationNode n) {
		connectPoints[1].removeLoopOrigin(n);
	}

	@Override
	public boolean checkLoops(PreparationNode origin) {
		return connectPoints[1].checkLoops(origin);
	}
	
	@Override
	public void execute(Program p) {
		if (hasBreakpoint && p.isRunning()) p.pause();
		else {
			p.addVariable(operation.getFirstOperand());
			Value val = operation.getSecondOperand().getValue();
			p.setValue(operation.getFirstOperand(), val);
			p.updateMonitor(operation.getFirstOperand());
			connectPoints[1].nextStep(p);
		}
	}
	
	@Override
	public void save(XMLWriter h) throws XMLWriterException {
		h.openNodeElem(Const.PROCESS_CODE, positionX, positionY);
		h.saveConnectPoints(connectPoints);
		operation.save(h);
		h.closeElem();
	}

	@Override
	public void generateCCode(CGenerator g) {
		g.newLine(operation.getFirstOperand().print() + Const.SPACE + Const.ASSIGN_OPERATION_SIGN + Const.SPACE + operation.getSecondOperand().print() + Const.SEMICOLON);
		connectPoints[1].generateCCode(g);
	}

	@Override
	public void collectVariables(LinkedList<Variable> vars) {
		operation.collectVariables(vars);
		connectPoints[1].collectVariables(vars);
	}

}
