package jflowchart.engine.runtime;

import java.util.LinkedList;

import jflowchart.engine.arithmetics.IncompatibleDataTypesException;
import jflowchart.engine.arithmetics.Value;
import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.codegen.c.CGenerator;
import jflowchart.engine.nodes.Node;
import jflowchart.engine.nodes.StartNode;
import jflowchart.ui.canvas.FlowChartCanvas;

enum ProgramState {
	RUNNING, PAUSED, IDLE
}

public class Program {
	
	private StartNode start;
	private Node currentNode;
	private LinkedList<Variable> environment;
	private Variable unsetVariable;
	private ProgramState state;
	private FlowChartCanvas canvas;
	
	public Program(FlowChartCanvas c) {
		discardStart();
		canvas = c;
		environment = new LinkedList<Variable>();
		unsetVariable = null;
		state = ProgramState.IDLE;
	}
	
	public void setStart(StartNode s) {
		start = s;
		currentNode = s;
	}
	
	public StartNode getStart() {
		return start;
	}
	
	public void discardStart() {
		start = null;
		currentNode = null;
	}
	
	public boolean hasStart() {
		return (start != null);
	}
	
	public Variable getUnsetVariable() {
		return unsetVariable;
	}
	
	public void addVariable(Variable v) {
		if (!environment.contains(v))
			environment.add(v);
	}
	
	public void setValue(Variable v, Value val) {
		if (environment.contains(v))
			try {
				v.setValue(val);
			} catch (IncompatibleDataTypesException e) {}
	}
	
	public void setUserSet(Variable v) {
		if (environment.contains(v))
			v.setUserSet();
	}
	
	public void setIterating(Variable v) {
		if (environment.contains(v))
			v.setIterating();
	}
	
	public boolean isSet(Variable v) {
		return environment.contains(v);
	}
	
	private void unsetEnvironment() {
		unsetVariable = null;
		for (Variable v : environment) 
			v.discardValue();
		environment.clear();
	}
	
	public boolean checkVariables() {
		unsetEnvironment();
		if (start == null) return true;
		else {
			unsetVariable = start.checkVariables(this);
			return (unsetVariable == null);
		}
	}
	
	public boolean isReachableEnd() {
		unsetEnvironment();
		if (start == null) return false;
		else return (start.isReachableEnd(this));
	}
	
	public boolean hasInfiniteLoop() {
		unsetEnvironment();
		if (start == null) return false;
		return start.hasInfiniteLoop(this, null);
	}
	
	public boolean isRunning() {
		return (state == ProgramState.RUNNING);
	}
	
	public boolean isPaused() {
		return (state == ProgramState.PAUSED);
	}
	
	public boolean isIdle() {
		return (state == ProgramState.IDLE);
	}
	
	public void setCurrentNode(Node n) {
		currentNode.setActive(false);
		currentNode = n;
		currentNode.setActive(true);
	}
	
	public void terminate() {
		state = ProgramState.IDLE;
		canvas.stopPlayer();
		unsetEnvironment();
		updateMonitor(null);
		canvas.unlock();
		currentNode.setActive(false);
		canvas.redraw();
		canvas.getMainShell().adjustNonSteppingItems(true);
		currentNode = start;
	}
	
	public void initRun() {
		unsetEnvironment();
		canvas.lock();
		currentNode.setActive(true);
		state = ProgramState.RUNNING;
	}
	
	public void pause() {
		state = ProgramState.PAUSED;
		canvas.stopPlayer();
	}
	
	public void nextStep() {
		currentNode.execute(this);
	}
	
	public FlowChartCanvas getCanvas() {
		return canvas;
	}
	
	public void updateMonitor(Variable lastChanged) {
		canvas.getMainShell().updateVariableMonitor(environment, lastChanged);
	}
	
	public LinkedList<Variable> collectVariables() {
		LinkedList<Variable> vars = new LinkedList<Variable>();
		start.collectVariables(vars);
		return vars;
	}
	
	public void generateCCode(CGenerator g) {
		start.generateCCode(g);
	}

}
