package jflowchart.engine.xml;

import java.io.InputStream;
import java.util.LinkedList;
import java.util.Stack;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamReader;

import jflowchart.engine.arithmetics.Assignment;
import jflowchart.engine.arithmetics.ComparableOperand;
import jflowchart.engine.arithmetics.Comparison;
import jflowchart.engine.arithmetics.IncompatibleDataTypesException;
import jflowchart.engine.arithmetics.InvalidOperatorException;
import jflowchart.engine.arithmetics.NumberExpression;
import jflowchart.engine.arithmetics.NumberOperand;
import jflowchart.engine.arithmetics.Operand;
import jflowchart.engine.arithmetics.Value;
import jflowchart.engine.arithmetics.Variable;
import jflowchart.engine.common.Const;
import jflowchart.engine.common.Functions;
import jflowchart.engine.nodes.AssistantConnectPoint;
import jflowchart.engine.nodes.ConnectPoint;
import jflowchart.engine.nodes.Connection;
import jflowchart.engine.nodes.InputOutputNode;
import jflowchart.ui.canvas.FlowChartCanvas;
import jflowchart.ui.canvas.graphnodes.DecisionGraphNode;
import jflowchart.ui.canvas.graphnodes.DisplayGraphNode;
import jflowchart.ui.canvas.graphnodes.EndGraphNode;
import jflowchart.ui.canvas.graphnodes.GraphNode;
import jflowchart.ui.canvas.graphnodes.ManualInputGraphNode;
import jflowchart.ui.canvas.graphnodes.PreparationGraphNode;
import jflowchart.ui.canvas.graphnodes.ProcessGraphNode;
import jflowchart.ui.canvas.graphnodes.StartGraphNode;

public class XMLReader {
	
	private XMLInputFactory xmlInputFactory;
	private XMLStreamReader reader;
	
	private int[] canvasSize;
	private FlowChartCanvas canvas;
	
	private GraphNode activeNode;
	private InputOutputNode activeIONode;
	private ProcessGraphNode activeProcessNode;
	private DecisionGraphNode activeDecisionNode;
	private PreparationGraphNode activePreparationNode;
	private LinkedList<GraphNode> nodes;
	
	private boolean collectingCPIDs;
	private LinkedList<String> connectPointIDs;
	
	private String variableLabel;
	private String variableType;
	private LinkedList<Variable> variables;
	
	private String valueFlag;
	
	private Assignment activeAssignment;
	private Comparison activeComparison;
	private Stack<NumberExpression> numberOps;
	
	private Connection activeConnection;
	private LinkedList<Connection> connections;
	private String pathX;
	private String pathY;
	
	public XMLReader() {
		
		xmlInputFactory = XMLInputFactory.newInstance();
		reader = null;
		
		canvasSize = new int[2];
		canvas = null;
		
		activeNode = null;
		activeIONode = null;
		activeProcessNode = null;
		activeDecisionNode = null;
		activePreparationNode = null;
		nodes = new LinkedList<GraphNode>();
		
		collectingCPIDs = false;
		connectPointIDs = new LinkedList<String>();
		
		variableLabel = null;
		variableType = null;
		variables = new LinkedList<Variable>();
		
		valueFlag = null;
		
		activeAssignment = null;
		activeComparison = null;
		numberOps = new Stack<NumberExpression>();
		
		activeConnection = null;
		connections = new LinkedList<Connection>();
		pathX = null;
		pathY = null;
		
	}
	
	private void setOperand(Operand o) throws XMLReaderException {
		
		if (!numberOps.empty()) {
			NumberExpression e = numberOps.peek();
			try {
				if (e.getFirstOperand() == null)
					e.setFirstOperand((NumberOperand)o);
				else {
					e.setSecondOperand((NumberOperand)o);
					e = numberOps.pop();
					setOperand(e);
				}
			} catch (IncompatibleDataTypesException e1) {
				throw new XMLReaderException("Incompatible data types");
			}
		} else if (activeAssignment != null) {
			try {
				if (activeAssignment.getSecondOperand() == null) {
					activeAssignment.setSecondOperand(o);
					activeAssignment = null;
				} else throw new XMLReaderException("Invalid placement of operand");
			} catch (IncompatibleDataTypesException e) {
				throw new XMLReaderException("Incompatible data types");
			}
		} else if (activeComparison != null) {
			try {
				if (activeComparison.getFirstOperand() == null)
					activeComparison.setFirstOperand((ComparableOperand)o);
				else {
					activeComparison.setSecondOperand((ComparableOperand)o);
					activeComparison = null;
				}
			} catch (IncompatibleDataTypesException e) {
				throw new XMLReaderException("Incompatible data types");
			}
		} else throw new XMLReaderException("Invalid placement of operand");
		
	}
	
	private void setVariable(Variable v) throws XMLReaderException {
		
		if (!numberOps.empty()) setOperand(v);
		else if (activeAssignment != null) {
			try {
				if (activeAssignment.getFirstOperand() == null)
					activeAssignment.setFirstOperand(v);
				else {
					activeAssignment.setSecondOperand(v);
					activeAssignment = null;
				}
			} catch (IncompatibleDataTypesException e) {
				throw new XMLReaderException("Incompatible data types");
			}
		} else if (activeComparison != null)
			setOperand(v);
		else if (activeIONode != null)
			activeIONode.setVariable(v);
		else throw new XMLReaderException("Invalid placement of variable " + v.getName());
		
	}
	
	private void newVariable(String label, String type, String name) throws XMLReaderException {
		
		Variable v = null;
		
		if (label.equals(Const.COMPARABLE_VARIABLE_CODE))
			v = Functions.parseComparableVariable(type, name, variables);
		else if (label.equals(Const.NUMBER_VARIABLE_CODE))
			v = Functions.parseNumberVariable(type, name, variables);
		
		if (v == null) throw new XMLReaderException("Failed to build variable: " + name);
		else setVariable(v);
		
		if (Functions.findVariable(v.getDataType(), name, variables) == null)
			variables.add(v);
		
	}
	
	private void newValue(String label, String value) throws XMLReaderException {
		
		Value v = null;
		
		if (label.equals(Const.COMPARABLE_VALUE_CODE))
			v = Functions.parseComparableValue(value);
		else if (label.equals(Const.NUMBER_VALUE_CODE))
			v = Functions.parseNumberValue(value);
		
		if (v == null) throw new XMLReaderException("Failed to build value: " + value);
		else setOperand(v);
		
	}
	
	private void newNumberExpression(String sign) throws XMLReaderException {
		NumberExpression e = new NumberExpression();
		try {
			e.setOperator(Functions.parseNumberOperatorCode(sign));
		} catch (InvalidOperatorException e1) {
			throw new XMLReaderException("Unknown number operation sign: " + sign);
		}
		numberOps.push(e);
	}
	
	private void setAssignment() throws XMLReaderException {
		if (activeProcessNode != null)
			activeAssignment = activeProcessNode.getOperation();
		else if (activePreparationNode != null)
			activeAssignment = activePreparationNode.getOperation();
		else throw new XMLReaderException("Invalid placement of Assignment node");
	}
	
	private void setComparison(String sign) throws XMLReaderException {
		
		if (activeDecisionNode != null)
			activeComparison = (Comparison)activeDecisionNode.getOperation();
		else if (activePreparationNode != null)
			activeComparison = activePreparationNode.getCondition();
		else throw new XMLReaderException("Invalid placement of Preparation node");
		
		try {
			activeComparison.setOperator(Functions.parseComparisonOperatorCode(sign));
		} catch (InvalidOperatorException e) {
			throw new XMLReaderException("Unknown comparison sign: " + sign);
		}
		
	}
	
	private void newNode(String label, String x, String y) throws XMLReaderException {
		
		if (activeNode != null)
			activeNode.setConnectPointIDs(connectPointIDs);
		
		if ((activeProcessNode != null) || (activePreparationNode != null))
			activeAssignment = null;
		
		if ((activeDecisionNode != null) || (activePreparationNode != null))
			activeComparison = null;
		
		int posX = Integer.parseInt(x);
		int posY = Integer.parseInt(y);
		
		GraphNode n = null;
		
		if ((posX > canvasSize[0]) || (posY > canvasSize[1]))
			throw new XMLReaderException(label + " [" + x + ", " + y + "] out of canvas bounds [" + canvasSize[0] + ", " + canvasSize[1] + "]");
		else {
			
			activeIONode = null;
			activeProcessNode = null;
			activeDecisionNode = null;
			activePreparationNode = null;
			
			if (label.equals(Const.START_NODE_CODE))
				n = new StartGraphNode(canvas, posX, posY, false);
			else if (label.equals(Const.END_NODE_CODE))
				n = new EndGraphNode(canvas, posX, posY, false);
			else if (label.equals(Const.PROCESS_CODE)) {
				n = new ProcessGraphNode(canvas, posX, posY, false);
				activeProcessNode = (ProcessGraphNode)n;
			} else if (label.equals(Const.MANUAL_INPUT_CODE)) {
				n = new ManualInputGraphNode(canvas, posX, posY, false);
				activeIONode = (InputOutputNode)n;
			} else if (label.equals(Const.DISPLAY_CODE)) {
				n = new DisplayGraphNode(canvas, posX, posY, false);
				activeIONode = (InputOutputNode)n;
			} else if (label.equals(Const.DECISION_CODE)) {
				n = new DecisionGraphNode(canvas, posX, posY, false);
				activeDecisionNode = (DecisionGraphNode)n;
			} else if (label.equals(Const.PREPARATION_CODE)) {
				n = new PreparationGraphNode(canvas, posX, posY, false);
				activePreparationNode = (PreparationGraphNode)n;
			} else throw new XMLReaderException("Unknown node label: " + label);
				
			nodes.add(n);
			activeNode = n;
		
		}
	
	}
	
	private ConnectPoint findConnectPoint(String id) throws XMLReaderException {
		ConnectPoint cp;
		for (GraphNode n : nodes) {
			cp = n.getConnectPointByID(id);
			if (cp != null) return cp;
		}
		throw new XMLReaderException("ConnectPoint ID " + id + " wasn't found");
	}
	
	private void newConnection(String from, String to) throws XMLReaderException {
		
		if (activeConnection != null)
			connections.add(activeConnection);
		
		ConnectPoint cpFrom = findConnectPoint(from);
		ConnectPoint cpTo = findConnectPoint(to);
		
		if (cpFrom.isInput() || (!cpTo.isInput()))
			throw new XMLReaderException("Invalid connection from " + from + " to " + to);
		else activeConnection = new Connection(cpFrom, cpTo);
	
	}
	
	private void addAssistantConnectPoint(String x, String y, String id) throws XMLReaderException {
		int posX = Integer.parseInt(x);
		int posY = Integer.parseInt(y);
		if ((posX > canvasSize[0]) || (posY > canvasSize[1]))
			throw new XMLReaderException("Assistant connect point [" + x + ", " + "] out of bounds");
		else {
			AssistantConnectPoint a = new AssistantConnectPoint(posX, posY, activeConnection.getLastOutput());
			a.setID(id);
			activeConnection.addPathPoint(a);
		}
	}
	
	private void readElem() throws XMLReaderException {
		
		String label = reader.getName().toString();
		
		collectingCPIDs = false;
		variableLabel = null;
		variableType = null;
		valueFlag = null;
		pathX = null;
		pathY = null;
		
		if (label.equals(Labels.CANVAS)) {
			canvasSize[0] = Integer.parseInt(reader.getAttributeValue(null, Labels.CANVAS_SIZE_X));
			canvasSize[1] = Integer.parseInt(reader.getAttributeValue(null, Labels.CANVAS_SIZE_Y));
		} else if (label.equals(Const.START_NODE_CODE) || label.equals(Const.END_NODE_CODE)
				|| label.equals(Const.PROCESS_CODE) || label.equals(Const.MANUAL_INPUT_CODE)
				|| label.equals(Const.DISPLAY_CODE) || label.equals(Const.DECISION_CODE)
				|| label.equals(Const.PREPARATION_CODE))
			newNode(label, reader.getAttributeValue(null, Labels.X), reader.getAttributeValue(null, Labels.Y));
		else if (label.equals(Labels.CONNECT_POINTS))
			connectPointIDs.clear();
		else if (label.equals(Labels.CONNECT_POINT))
			collectingCPIDs = true;
		else if (label.equals(Const.COMPARABLE_VARIABLE_CODE) || label.equals(Const.NUMBER_VARIABLE_CODE)) {
			variableLabel = label;
			variableType = reader.getAttributeValue(null, Labels.TYPE);
		} else if (label.equals(Const.ASSIGNMENT_CODE))
			setAssignment();
		else if (label.equals(Const.COMPARISON_CODE))
			setComparison(reader.getAttributeValue(null, Labels.OPERATOR));
		else if (label.equals(Const.COMPARABLE_VALUE_CODE) || label.equals(Const.NUMBER_VALUE_CODE))
			valueFlag = label;
		else if (label.equals(Const.NUMBER_EXPRESSION_CODE))
			newNumberExpression(reader.getAttributeValue(null, Labels.OPERATOR));
		else if (label.equals(Labels.CONNECTION))
			newConnection(reader.getAttributeValue(null, Labels.FROM), reader.getAttributeValue(null, Labels.TO));
		else if (label.equals(Labels.ASSISTANT_CONNECT_POINT)) {
			pathX = reader.getAttributeValue(null, Labels.X);
			pathY = reader.getAttributeValue(null, Labels.Y);
		} else if (label.equals(Labels.CONNECTIONS)) {
			if (activeNode != null)
				activeNode.setConnectPointIDs(connectPointIDs);
		}
	}
	
	private void setText() throws XMLReaderException {
		String txt = reader.getText();
		if (collectingCPIDs)
			connectPointIDs.add(txt);
		else if ((variableLabel != null) && (variableType != null))
			newVariable(variableLabel, variableType, txt);
		else if (valueFlag != null)
			newValue(valueFlag, txt);
		else if ((pathX != null) && (pathY != null))
			addAssistantConnectPoint(pathX, pathY, txt);
	}
	
	public void load(InputStream in, FlowChartCanvas c) throws XMLReaderException {
		
		try {
			
			reader = xmlInputFactory.createXMLStreamReader(in);
			canvas = c;
			
			canvasSize[0] = -1;
			canvasSize[1] = -1;
			
			nodes.clear();
			connectPointIDs.clear();
			
			numberOps.clear();
			
			connections.clear();
			
			while (reader.hasNext()) {
				switch (reader.next()) {
					case XMLStreamReader.START_ELEMENT:
						readElem();
						break;
					case XMLStreamReader.CHARACTERS:
						setText();
						break;
				}
			}
			
			canvas.setSize(canvasSize[0], canvasSize[1]);
			
			if (activeConnection != null)
				connections.add(activeConnection);
			
			for (Variable v : variables)
				canvas.addVariable(v);
			
			for (GraphNode n : nodes)
				canvas.addNode(n);
			
			for (Connection cn : connections)
				cn.addTo(canvas);
			
			
			
		} catch (Exception e) {
			throw new XMLReaderException(e.getMessage());
		}
		
	}

}
